/*
** File: legacy_pci.c  
** Project: VxWorks Driver Framework
** Purpose: Legacy (non-VxBus) entry points
**
** (C) Copyright Alpha Data 2010
*/

#include "df.h"
#include "dfvxworks.h"

#if !defined(ADB3_VXBUS)

static DfDriverObject g_dfDrvObj;
static boolean_t g_bPciDriverRegistered = FALSE;

static DfPciCompatibleId* g_pPciCompatibleIdTable = NULL;
static unsigned int g_numPciCompatibleId = 0;

static void
getRawBusResources(
  DfBusResources* pResources,
  int bus,
  int slot,
  int func)
{
  static const unsigned long barAddresses[] = {
    PCI_CFG_BASE_ADDRESS_0,
    PCI_CFG_BASE_ADDRESS_1,
    PCI_CFG_BASE_ADDRESS_2,
    PCI_CFG_BASE_ADDRESS_3,
    PCI_CFG_BASE_ADDRESS_4,
    PCI_CFG_BASE_ADDRESS_5
  };
  unsigned int i;
  unsigned int ioBarCount, memBarCount;
  boolean_t b64Bit, bIo;
  UINT32 currLo, currHi;
  UINT32 maskLo, maskHi;
  uint8_t flags;
  uint8_t intLine, intPin;
  uint64_t mask64;

  /* Get the BAR sizes */
  ioBarCount = 0;
  memBarCount = 0;
  for (i = 0; i < DF_ARRAY_LENGTH(barAddresses); i++) {
    flags = 0;
    pciConfigInLong(bus, slot, func, barAddresses[i], &currLo);
    if (currLo & PCI_BASE_IO) {
      currLo &= PCI_IOBASE_MASK;
      b64Bit = FALSE;
      bIo = TRUE;
    } else {
      if (currLo & PCI_BASE_IN_64BITS) {
        b64Bit = TRUE;
      } else {
        b64Bit = FALSE;
      }
      if (currLo & PCI_BASE_PREFETCH) {
        flags |= DF_PCIBARF_PREFETCH;
      }
      flags = (uint8_t)(currLo & ~PCI_MEMBASE_MASK);
      flags &= 0xFEU;
      currLo &= PCI_MEMBASE_MASK;
      bIo = FALSE;
    }
    if (b64Bit) {
      /* 64-bit BAR */
      if (i == 5) {
        continue;
      }
      pciConfigInLong(bus, slot, func, barAddresses[i + 1], &currHi);
      pciConfigOutLong(bus, slot, func, barAddresses[i], ~0);
      pciConfigOutLong(bus, slot, func, barAddresses[i + 1], ~0);
      pciConfigInLong(bus, slot, func, barAddresses[i], &maskLo);
      pciConfigInLong(bus, slot, func, barAddresses[i + 1], &maskHi);
      maskLo &= PCI_MEMBASE_MASK;
      pciConfigOutLong(bus, slot, func, barAddresses[i], currLo);
      pciConfigOutLong(bus, slot, func, barAddresses[i + 1], currHi);
      if ((maskLo == 0 && maskHi == 0) || (maskLo != 0 && (b64Bit && maskHi != 0xffffffffUL))) {
          /* BAR size is 0 (or invalid) */
      } else {
        mask64 = (uint64_t)maskLo;
        mask64 |= ((uint64_t)maskHi) << 32;
        dfAssert(mask64 != 0xffffffffffffffffULL);
        pResources->variant.pci.memBar[memBarCount].size = ~mask64 + 1;
        pResources->variant.pci.memBar[memBarCount].base = (uint64_t)currLo;
        pResources->variant.pci.memBar[memBarCount].base |= ((uint64_t)currHi) << 32;
        pResources->variant.pci.memBar[memBarCount].flags = flags;
        memBarCount++;
      }
      i++;
    } else {
      /* 32-bit BAR */
      pciConfigOutLong(bus, slot, func, barAddresses[i], ~0);
      pciConfigInLong(bus, slot, func, barAddresses[i], &maskLo);
      maskLo &= bIo ? PCI_IOBASE_MASK : PCI_MEMBASE_MASK;
      pciConfigOutLong(bus, slot, func, barAddresses[i], currLo);
      if (maskLo == 0) {
        /* BAR size is 0 */
      } else {
        dfAssert(maskLo != 0xffffffffU);
        if (bIo) {
          pResources->variant.pci.ioBar[ioBarCount].base = (uint64_t)currLo;
          pResources->variant.pci.ioBar[ioBarCount].size = (uint64_t)(~maskLo + 1);
          pResources->variant.pci.ioBar[ioBarCount].flags = flags;
          ioBarCount++;
        } else {
          pResources->variant.pci.memBar[memBarCount].base = (uint64_t)currLo;
          pResources->variant.pci.memBar[memBarCount].size = (uint64_t)(~maskLo + 1);
          pResources->variant.pci.memBar[memBarCount].flags = flags;
          memBarCount++;
        }
      }
    }
  }
  pResources->variant.pci.numMemBar = memBarCount;
  pResources->variant.pci.numIoBar = ioBarCount;

  pciConfigInByte(bus, slot, func, PCI_CFG_DEV_INT_PIN, &intPin);
  if (intPin == 0) {
    pResources->variant.pci.numIrq = 0;
  } else {
    pciConfigInByte(bus, slot, func, PCI_CFG_DEV_INT_LINE, &intLine);
    pResources->variant.pci.numIrq = 1;
    pResources->variant.pci.irq[0].bMessageSignalled = FALSE;
    pResources->variant.pci.irq[0].bLevelSensitive = TRUE;
    pResources->variant.pci.irq[0].variant.raw.intPin = intPin;
    pResources->variant.pci.irq[0].variant.raw.intLine = intLine;
  }
}

static STATUS
pciProbe(
  DfDriverObject* pDrvObj,
  DfPciCompatibleId* pPciCompatibleId,
  int bus,
  int slot,
  int func)
{
  STATUS status = OK;
  DfHardwareId hardwareId;
  DfBusResources rawResources;
  DfDeviceObject* pDevObj = NULL;
  uint16_t subVendor, subDevice, cmdReg, newCmdReg;
  uint8_t revision;
  boolean_t bStarted = FALSE, bAdded = FALSE;

  pciConfigInWord(bus, slot, func, PCI_CFG_SUB_VENDER_ID, &subVendor);
  pciConfigInWord(bus, slot, func, PCI_CFG_SUB_SYSTEM_ID, &subDevice);
  pciConfigInByte(bus, slot, func, PCI_CFG_REVISION, &revision);
  if ((!(pPciCompatibleId->match & DF_PCI_MATCH_SUBVENDOR) || subVendor == pPciCompatibleId->id.subVendor) &&
    (!(pPciCompatibleId->match & DF_PCI_MATCH_SUBDEVICE) || subDevice == pPciCompatibleId->id.subDevice) &&
    (!(pPciCompatibleId->match & DF_PCI_MATCH_REVISION) || revision == pPciCompatibleId->id.revision))
  {
    hardwareId.busType = DfBusPci;
    hardwareId.variant.pci.vendor = pPciCompatibleId->id.vendor;
    hardwareId.variant.pci.device = pPciCompatibleId->id.device;
    hardwareId.variant.pci.subVendor = subVendor;
    hardwareId.variant.pci.subDevice = subDevice;
    hardwareId.variant.pci.revision = revision;
    getRawBusResources(&rawResources, bus, slot, func);

    pDevObj = (DfDeviceObject*)dfMalloc(sizeof(*pDevObj));
    if (pDevObj == NULL) {
      status = ENOMEM;
      dfDebugPrint(0, ("*** pciProbe: Cannot allocate device object\n"));
      goto done;
    }
    pDevObj->system.pContext = NULL;
    pDevObj->system.pExportedInterfaces = NULL;
    pDevObj->system.pImportedInterfaces = NULL;
    pDevObj->system.pMmappableRegions = NULL;
    pDevObj->system.interfaceList.pHead = NULL;
    pDevObj->system.dma.pMapArray = NULL;
    pDevObj->system.request.bPoolValid = FALSE;
    pDevObj->system.request.bPoolNonBlockValid = FALSE;
    pDevObj->system.pNext = NULL;
    pDevObj->hardwareId = hardwareId;

    if (!dfPoolInit(&pDevObj->system.request.pool, pDevObj, "clReq", DfClientRequest)) {
      status = ENOMEM;
      dfDebugPrint(0, ("*** pciProbe: Cannot init I/O request pool\n"));
      goto done;
    }
    pDevObj->system.request.bPoolValid = TRUE;

    if (!dfPoolInit(&pDevObj->system.request.poolNonBlock, pDevObj, "clReqNB", DfVxwClientRequestNonBlock)) {
      status = ENOMEM;
      dfDebugPrint(0, ("*** pciProbe: Cannot init non-blocking I/O request pool\n"));
      goto done;
    }
    pDevObj->system.request.bPoolNonBlockValid = TRUE;

    if (g_dfDrvObj.deviceContextSize != 0) {
      pDevObj->system.pContext = dfMalloc(g_dfDrvObj.deviceContextSize);
      if (NULL == pDevObj->system.pContext) {
        status = ENOMEM;
        dfDebugPrint(0, ("*** pciProbe: Cannot allocate device context, size=%lu(0x%lx) bytes\n",
          (unsigned long)g_dfDrvObj.deviceContextSize, (unsigned long)g_dfDrvObj.deviceContextSize));
        goto done;
      }
    }

    if (NULL != g_dfDrvObj.methods.pOnAdd) {
      dfDebugPrint(1, ("pciProbe: calling OnAdd method\n"));
      if (!g_dfDrvObj.methods.pOnAdd(pDevObj)) {
        status = ENODEV;
        dfDebugPrint(0, ("*** pciProbe: OnAdd failed\n"));
        goto done;
      }
      bAdded = TRUE;
    }

    pciConfigInWord(bus, slot, func, PCI_CFG_COMMAND, &cmdReg);
    newCmdReg = cmdReg;
    if (!(cmdReg & PCI_CMD_MASTER_ENABLE)) {
      dfDebugPrint(1, ("pciProbe: enabling bus master\n"));
      newCmdReg |= PCI_CMD_MASTER_ENABLE;
    }
    if (!(cmdReg & PCI_CMD_MEM_ENABLE)) {
      dfDebugPrint(1, ("pciProbe: enabling memory response\n"));
      newCmdReg |= PCI_CMD_MEM_ENABLE;
    }
    if (cmdReg != newCmdReg) {
      pciConfigOutWord(bus, slot, func, PCI_CFG_COMMAND, newCmdReg);
    }

    if (NULL != g_dfDrvObj.methods.pOnStart) {
      dfDebugPrint(1, ("pciProbe: calling OnStart method\n"));
      if (!g_dfDrvObj.methods.pOnStart(pDevObj, &rawResources, &rawResources)) { /* raw and xlat resources are the same */
        status = ENODEV;
        dfDebugPrint(0, ("*** pciProbe: OnStart failed\n"));
        goto done;
      }
      bStarted = TRUE;
    }

    /* Add this device to the driver's list */
    if (g_dfDrvObj.system.devices.pHead == NULL) {
      g_dfDrvObj.system.devices.pHead = pDevObj;
    } else {
      g_dfDrvObj.system.devices.pHead->system.pNext = pDevObj;
    }
    g_dfDrvObj.system.devices.pTail = pDevObj;

done:
    if (OK != status) {
      if (NULL != pDevObj) {
        if (bStarted) {
          g_dfDrvObj.methods.pOnStop(pDevObj);
          bStarted = FALSE;
        }
        if (bAdded) {
          g_dfDrvObj.methods.pOnRemove(pDevObj);
          bAdded = FALSE;
        }
        if (pDevObj->system.request.bPoolNonBlockValid) {
          dfPoolUninit(&pDevObj->system.request.poolNonBlock);	
          pDevObj->system.request.bPoolNonBlockValid = FALSE;
        }
        if (pDevObj->system.request.bPoolValid) {
          dfPoolUninit(&pDevObj->system.request.pool);	
          pDevObj->system.request.bPoolValid = FALSE;
        }
        if (NULL != pDevObj->system.pContext) {
          dfFree(pDevObj->system.pContext);
          pDevObj->system.pContext = NULL;
        }
        dfFree(pDevObj);
        pDevObj = NULL;
      }
    }
  }
  return status;
}

static STATUS
removeDevice(
  DfDeviceObject* pDevObj)
{
  if (NULL != g_dfDrvObj.methods.pOnStop) {
    dfDebugPrint(1, ("removeDevice: calling OnStop method\n"));
    g_dfDrvObj.methods.pOnStop(pDevObj);
  }
  
  if (NULL != g_dfDrvObj.methods.pOnRemove) {
    dfDebugPrint(1, ("removeDevice: calling OnRemove method\n"));
    g_dfDrvObj.methods.pOnRemove(pDevObj);
  }

  if (pDevObj->system.request.bPoolNonBlockValid) {
    dfPoolUninit(&pDevObj->system.request.poolNonBlock);	
    pDevObj->system.request.bPoolNonBlockValid = FALSE;
  }
  if (pDevObj->system.request.bPoolValid) {
    dfPoolUninit(&pDevObj->system.request.pool);	
    pDevObj->system.request.bPoolValid = FALSE;
  }

  if (NULL != pDevObj->system.pContext && g_dfDrvObj.deviceContextSize != 0) {
    dfFree(pDevObj->system.pContext);
  }
  dfFree(pDevObj);

  return OK;
}

static STATUS
registerDriver(
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  STATUS err = OK;
  DfPciCompatibleId* pPciCompatibleId;
  unsigned int i, instance, nId, nLegacyId;
  int bus, slot, func;

  /* Count the number of PCI hardware IDs */
  for (nId = 0; pPciCompatibleIdTable[nId].match != 0xffU; nId++);
  g_numPciCompatibleId = nId;

  if (bLegacyHardware) {
    /* Count the number of legacy PCI hardware IDs */
    for (nLegacyId = 0; pLegacyPciCompatibleIdTable[nLegacyId].match != 0xffU; nLegacyId++);
    g_numPciCompatibleId += nLegacyId;
  } else {
    nLegacyId = 0;
  }

  /* Make our own compatible PCI ID table in case the ones passed in become invalid later */
  g_pPciCompatibleIdTable = (DfPciCompatibleId*)dfMalloc(sizeof(DfPciCompatibleId) * g_numPciCompatibleId);
  if (NULL == g_pPciCompatibleIdTable) {
    dfDebugPrint(0, ("registerDriver: failed to allocate PCI Compatible ID table\n"));
    err = ERROR;
    goto done;
  }
  bcopy((const char*)pPciCompatibleIdTable, (char*)g_pPciCompatibleIdTable, sizeof(DfPciCompatibleId) * nId);
  if (bLegacyHardware) {
    bcopy((const char*)pLegacyPciCompatibleIdTable, (char*)(g_pPciCompatibleIdTable +  nId), sizeof(DfPciCompatibleId) * nLegacyId);
  }

  for (i = 0; i < g_numPciCompatibleId; i++) {
    instance = 0;
    pPciCompatibleId = &g_pPciCompatibleIdTable[i];
    dfAssert((pPciCompatibleId->match & DF_PCI_MATCH_VENDOR) && (pPciCompatibleId->match & DF_PCI_MATCH_DEVICE));
    while (1) {
      if (OK == pciFindDevice((int)pPciCompatibleId->id.vendor, (int)pPciCompatibleId->id.device, instance, &bus, &slot, &func)) {
        pciProbe(&g_dfDrvObj, pPciCompatibleId, bus, slot, func);
        instance++;
      } else {
        break;
      }
    }
  }

done:
  if (OK != err) {
    if (NULL != g_pPciCompatibleIdTable) {
      dfFree(g_pPciCompatibleIdTable);
      g_pPciCompatibleIdTable = NULL;
    }
  } else {
    g_bPciDriverRegistered = TRUE;
  }
  return err;
}

static void
unregisterDriver(
  void)
{
  DfDeviceObject* pDevObj;

  /* Delete all of the devices in the driver's list */
  while (NULL != g_dfDrvObj.system.devices.pHead) {
    pDevObj = g_dfDrvObj.system.devices.pHead;
    g_dfDrvObj.system.devices.pHead = pDevObj->system.pNext;
    removeDevice(pDevObj);
  }
  g_dfDrvObj.system.devices.pTail = NULL;

  if (NULL != g_pPciCompatibleIdTable) {
    dfFree((char*)g_pPciCompatibleIdTable);
    g_pPciCompatibleIdTable = NULL;
  }

  g_bPciDriverRegistered = FALSE;
}

/*
** ----------------------------------------------------------------------
** Functions for starting and stopping the driver
** ----------------------------------------------------------------------
*/

STATUS
dfVxwLegacyEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  STATUS err = OK;

#if DF_DBG_BUILD
  g_dfDebugLevel = debugLevel;
#endif
  dfVxwLogInit();

  dfDebugPrint(1, ("dfVxwLegacyEntry: entered\n"));

  dfVxwCheckDataTypes();

  if (!dfVxwSemPoolInit()) {
    dfDebugPrint(0, ("*** dfVxbEntry: failed to initialize binary semaphore pool\n"));
    dfVxwLogUninit();
    return ERROR;
  }

  DF_ZERO_OBJECT(&g_dfDrvObj);
  g_dfDrvObj.methods.pOnAdd = NULL;
  g_dfDrvObj.methods.pOnRemove = NULL;
  g_dfDrvObj.methods.pOnStart = NULL;
  g_dfDrvObj.methods.pOnStop = NULL;
  g_dfDrvObj.methods.pOnPowerQuery = NULL;
  g_dfDrvObj.methods.pOnPowerSet = NULL;
  g_dfDrvObj.system.devices.pHead = NULL;
  g_dfDrvObj.system.devices.pTail = NULL;

  dfDebugPrint(1, ("dfVxwLegacyEntry: calling dfDriverEntry\n"));
  if (!dfDriverEntry(&g_dfDrvObj)) {
    dfVxwSemPoolUninit();
    dfVxwLogUninit();
    return ERROR;
  }

  dfDebugPrint(1, ("dfVxwLegacyEntry: searching for devices\n"));
  err = registerDriver(pPciCompatibleIdTable, pLegacyPciCompatibleIdTable, bLegacyHardware);
  if (OK != err) {
    dfDriverExit(&g_dfDrvObj);
    dfVxwSemPoolUninit();
    dfVxwLogUninit();
    return err;
  }

  return err;
}

void
dfVxwLegacyExit(
  void)
{
  dfDebugPrint(1, ("dfVxwLegacyExit: entered\n"));

  if (g_bPciDriverRegistered) {
    dfDebugPrint(1, ("dfVxwLegacyExit: deleting devices\n"));
    unregisterDriver();
  }

  dfDebugPrint(1, ("dfVxwLegacyEntry: calling dfDriverExit\n"));
  dfDriverExit(&g_dfDrvObj);

  dfVxwSemPoolUninit();
  dfVxwLogUninit();
}

#else

STATUS
dfVxwLegacyEntry(
  int debugLevel,
  const char* pDriverName,
  const DfPciCompatibleId* pPciCompatibleIdTable,
  const DfPciCompatibleId* pLegacyPciCompatibleIdTable,
  boolean_t bLegacyHardware)
{
  dfAssert(FALSE);
  return ERROR;
}

void
dfVxwLegacyExit(
  void)
{
  dfAssert(FALSE);
}

#endif
